import pytest

from configyaml.validator import ConfigValidator


def test_invalid_configvalidator_subclass():
    class InvalidConfigValidator(ConfigValidator):
        pass

    with pytest.raises(AttributeError):
        InvalidConfigValidator(config_text="foo: bar")



from configyaml.config import PositiveIntegerNode, StringNode, WildcardDictNode, BoolNode, RegexNode
from configyaml.loader import ConfigLoader


class DummyPosIntLoader(ConfigLoader):
    config_root_class = PositiveIntegerNode


def test_wildcard_dict_invalid_key():
    value = {'_bad': 'key'}
    c = WildcardDictNode(value=value)
    assert not c.is_valid()
    assert c._errors[0].title == 'Invalid field name'
    assert c._errors[0].description == 'Cannot start field name with a "_"'

    value = {'*': 'key'}
    c = WildcardDictNode(value=value)
    assert not c.is_valid()
    assert c._errors[0].title == 'Invalid field name'
    assert c._errors[0].description == 'Field name cannot be "*"'


def test_pos_int_invalid_validation():
    value = '-1'
    loader = DummyPosIntLoader(value)
    assert not loader.is_valid()
    assert loader.errors[0].title == "Invalid Value"


def test_pos_int_valid_validation():
    value = '1'
    loader = DummyPosIntLoader(value)
    assert loader.is_valid()


def test_pos_int_zero_validation():
    value = '0'
    loader = DummyPosIntLoader(value)
    assert loader.is_valid()


def test_str_valid():
    value = 'a simple string'
    loader = StringNode(value)
    assert loader.is_valid()


def test_str_invalid():
    value = 99
    loader = StringNode(value)
    assert not loader.is_valid()


# BoolNode tests
def test_bool_valid():
    value = True
    node = BoolNode(value=value)
    assert node.is_valid()

    value = False
    node = BoolNode(value=value)
    assert node.is_valid()


def test_bool_invalid():
    value = 'True'
    node = BoolNode(value=value)
    assert not node.is_valid()

    value = 1
    node = BoolNode(value=value)
    assert not node.is_valid()


# RegexNode tests
def test_regex_valid():
    value = '.*'
    node = RegexNode(value=value)
    assert node.is_valid()
    assert node.regex != None


def test_regex_invalid():
    value = '['
    node = RegexNode(value=value)
    assert not node.is_valid()
    assert node._errors[0].title == 'Invalid regex'
    assert node._errors[0].description in ('unexpected end of regular expression', 'unterminated character set at position 0')
    assert node.regex == None



from .base import AbstractNode


class ChoiceNode(AbstractNode):
    _choices = None

    def __init__(self, *args, **kwargs):
        self._type = str

        if not self._choices:
            # class should specify self._choices = []  # list of choices that the str can be
            raise AttributeError('_choices must be defined in subclasses of ConfigBaseChoices')

        super(ChoiceNode, self).__init__(*args, **kwargs)

    def _validate_value(self):
        if self._value not in self._choices:
            self._add_error(
                title='Invalid choice',
                description='Must be one of {}'.format(', '.join(self._choices)),
            )



from __future__ import unicode_literals


class ConfigError(object):
    def __init__(self, title, description=None, line=None, column=None, *args, **kwargs):
        self.title = title
        self.description = description

        # zero based
        self.line = line
        self.column = column
        self.start_line = self.line
        self.start_column = self.column
        self.end_line = kwargs.get('end_line', None)
        self.end_column = kwargs.get('end_column', None)

        self.error_obj = kwargs.get('error_obj', None)

    def __str__(self):
        return self.title

    def as_dict(self):
        return {
            'title': self.title,
            'description': self.description,
            'start_line': self.start_line,
            'start_column': self.start_column,
            'end_line': self.end_line,
            'end_column': self.end_column,
        }

    @classmethod
    def create_from_yaml_error(cls, yaml_error):
        if hasattr(yaml_error, 'problem_mark'):
            mark = yaml_error.problem_mark
            return cls(
                line=mark.line,
                column=mark.column,
                title='Basic YAML parsing error',
                description=yaml_error.problem,
                error_obj=yaml_error
            )
        else:
            return cls(
                line=None,
                column=None,
                title='Unknown YAML parsing error',
                description=yaml_error.problem,
                error_obj=yaml_error
            )

    @classmethod
    def create_from_yaml_node(cls, node, *args, **kwargs):
        if 'error_obj' not in kwargs:
            # if no error_obj specified, use node
            kwargs['error_obj'] = node

        if 'line' not in kwargs:
            kwargs['line'] = node.start_mark.line
            kwargs['end_line'] = node.end_mark.line

        if 'column' not in kwargs:
            kwargs['column'] = node.start_mark.column
            kwargs['end_column'] = node.end_mark.column

        return cls(*args, **kwargs)



import pytest

from configyaml.config.nodes import StringNode, BoolNode
from .test_loader import DummyLoader, DummyComplexLoader


def test_variable_match():
    n = StringNode(value='$var', variables={'var': 'testing'})
    assert n.is_valid()
    assert n._value == 'testing'

    assert n._as_dict() == {'value': 'testing'}
    assert n._as_dict(redact=True) == {'value': '[REDACTED]', 'redacted': True}


def test_variable_no_match():
    n = StringNode(value='$$var', variables={'var': 'testing'})
    assert not n.is_valid()
    assert n._value == '$$var'
    assert n._errors[0].title == 'Variable not found'

    assert n._as_dict() == {
        'errors': [
            {
                'description': "'$var' was not found in ['var']",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$$var',
    }

    assert n._as_dict(redact=True) == {
        'errors': [
            {
                'description': "'$var' was not found in ['var']",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$$var',
    }


def test_variable_escaped():
    n = StringNode(value='\$var', variables={'var': 'testing'})
    assert n.is_valid()
    assert n._value == '$var'

    assert n._as_dict() == {'value': '$var'}
    assert n._as_dict(redact=True) == {'value': '$var'}


def test_variable_case_sensitive():
    n = StringNode(value='$VAR', variables={'var': 'testing'})
    assert not n.is_valid()
    assert n._value == '$VAR'
    assert n._errors[0].title == 'Variable not found'

    assert n._as_dict() == {
        'errors': [
            {
                'description': "'VAR' was not found in ['var']",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$VAR',
    }

    assert n._as_dict(redact=True) == {
        'errors': [
            {
                'description': "'VAR' was not found in ['var']",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$VAR',
    }

    n = StringNode(value='$vaR', variables={'vaR': 'testing'})
    assert n.is_valid()
    assert n._value == 'testing'

    assert n._as_dict() == {'value': 'testing'}
    assert n._as_dict(redact=True) == {'value': '[REDACTED]', 'redacted': True}


def test_no_variables():
    n = StringNode(value='$var')
    assert not n.is_valid()
    assert n._value == '$var'

    assert n._as_dict() == {
        'errors': [
            {
                'description': "'var' was not found in []",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$var',
    }

    assert n._as_dict(redact=True) == {
        'errors': [
            {
                'description': "'var' was not found in []",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'Variable not found',
            }
        ],
        'value': '$var',
    }


def test_variables_type_error():
    with pytest.raises(TypeError):
        StringNode(value='$var', variables='testing')


def test_bool_node_variable():
    n = BoolNode(value='$test', variables={'test': True})
    assert n.is_valid()
    assert n._value == True

    assert n._as_dict() == {'value': True}
    assert n._as_dict(redact=True) == {'value': '[REDACTED]', 'redacted': True}


def test_bool_node_variable_bad_type():
    n = BoolNode(value='$test', variables={'test': 'ok'})
    assert not n.is_valid()
    assert n._value == 'ok'

    assert n._as_dict() == {
        'errors': [
            {
                'description': "boolnode must be a bool",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'boolnode has an invalid type',
            }
        ],
        'value': 'ok',
    }

    assert n._as_dict(redact=True) == {
        'errors': [
            {
                'description': "boolnode must be a bool",
                'end_column': None,
                'end_line': None,
                'start_column': None,
                'start_line': None,
                'title': 'boolnode has an invalid type',
            }
        ],
        'value': '[REDACTED]',
        'redacted': True,
    }


def test_loader_with_variables():
    value = "{ 'foo': '$bar'}"
    loader = DummyLoader(value, variables={'bar': 'testing'})
    assert loader.is_valid()

    assert loader.as_text() == "{ 'foo': '$bar'}"
    assert loader.as_dict() == {
        'config': {'foo': {'value': 'testing'}}, 'config_text': "{ 'foo': '$bar'}"
    }
    assert loader.as_dict(redact=True) == {
        'config': {'foo': {'redacted': True, 'value': '[REDACTED]'}},
        'config_text': "{ 'foo': '$bar'}",
    }


def test_loader_with_missing_variable():
    value = "{ 'foo': '$bar'}"
    loader = DummyLoader(value, variables={'boo': 'testing'})
    assert not loader.is_valid()
    assert loader.errors[0].title == 'Variable not found'

    assert loader.as_text() == """{ 'foo': '$bar'}
# { 'foo': '$bar'}
#     ^^^^^^
# --------
# Variable not found
# - 'bar' was not found in ['boo']
# --------"""

    assert loader.as_dict() == {
        'config': {
            'foo': {
                'errors': [
                    {
                        'description': "'bar' was not found in ['boo']",
                        'end_column': 15,
                        'end_line': 0,
                        'start_column': 9,
                        'start_line': 0,
                        'title': 'Variable not found',
                    }
                ],
                'value': '$bar',
            }
        },
        'config_text': "{ 'foo': '$bar'}",
        'errors': [
            {
                'description': "'bar' was not found in ['boo']",
                'end_column': 15,
                'end_line': 0,
                'start_column': 9,
                'start_line': 0,
                'title': 'Variable not found',
            }
        ],
    }
    assert loader.as_dict(redact=True) == {
        'config': {
            'foo': {
                'errors': [
                    {
                        'description': "'bar' was not found in ['boo']",
                        'end_column': 15,
                        'end_line': 0,
                        'start_column': 9,
                        'start_line': 0,
                        'title': 'Variable not found',
                    }
                ],
                'value': '$bar',
            }
        },
        'config_text': "{ 'foo': '$bar'}",
        'errors': [
            {
                'description': "'bar' was not found in ['boo']",
                'end_column': 15,
                'end_line': 0,
                'start_column': 9,
                'start_line': 0,
                'title': 'Variable not found',
            }
        ],
    }


def test_loader_with_nested_variables():
    value = 'foo: $list_var\nbar: yay'
    variables = {'list_var': ['one', 'two', '$three_var'], 'three_var': 'THREE!'}
    loader = DummyComplexLoader(value, variables=variables)
    assert loader.is_valid()

    assert loader.as_text() == 'foo: $list_var\nbar: yay'
    assert loader.as_dict() == {
        'config': {
            'bar': {'value': 'yay'},
            'foo': {'items': [{'value': 'one'}, {'value': 'two'}, {'value': 'THREE!'}]},
        },
        'config_text': 'foo: $list_var\nbar: yay',
    }
    assert loader.as_dict(redact=True) == {
        'config': {
            'bar': {'value': 'yay'}, 'foo': {'redacted': True, 'value': '[REDACTED]'}
        },
        'config_text': 'foo: $list_var\nbar: yay',
    }


def test_loader_with_nested_list_variable_missing():
    value = 'foo: $list_var\nbar: yay'
    variables = {
        'list_var': ['one', 'two', '$three_var'], 'threeve': 'THREE!'  # one key
    }
    loader = DummyComplexLoader(value, variables=variables)
    assert not loader.is_valid()

    text = loader.as_text()
    assert text == """foo: $list_var
# foo: $list_var
# ^^^^^^^^^
# --------
# Variable not found
# - 'three_var' was not found in ['list_var', 'threeve']
# --------
bar: yay"""

    # the variable content should not be visible, only the variable keys
    assert '$three_var' not in text

    assert loader.as_dict() == {
        'config': {
            'bar': {'value': 'yay'},
            'foo': {
                'items': [
                    {'value': 'one'},
                    {'value': 'two'},
                    {
                        'errors': [
                            {
                                'description': "'three_var' was not "
                                'found in '
                                "['list_var', "
                                "'threeve']",
                                'end_column': 14,
                                'end_line': 0,
                                'start_column': 5,
                                'start_line': 0,
                                'title': 'Variable not found',
                            }
                        ],
                        'value': '$three_var',
                    },
                ]
            },
        },
        'config_text': 'foo: $list_var\nbar: yay',
        'errors': [
            {
                'description': "'three_var' was not found in ['list_var', "
                "'threeve']",
                'end_column': 14,
                'end_line': 0,
                'start_column': 5,
                'start_line': 0,
                'title': 'Variable not found',
            }
        ],
    }
    assert loader.as_dict(redact=True) == {
        'config': {
            'bar': {'value': 'yay'}, 'foo': {'redacted': True, 'value': '[REDACTED]'}
        },
        'config_text': 'foo: $list_var\nbar: yay',
        'errors': [
            {
                'description': "'three_var' was not found in ['list_var', "
                "'threeve']",
                'end_column': 14,
                'end_line': 0,
                'start_column': 5,
                'start_line': 0,
                'title': 'Variable not found',
            }
        ],
    }



from .loader import ConfigLoader


class ConfigValidator(object):
    """A stripped down class for performing validation only (uses dummy context)"""
    configyaml_class = ConfigLoader
    context = None

    def __init__(self, config_text):
        if not self.context:
            raise AttributeError('self.context must be defined in subclasses of ConfigValidator')
        self.loader = self.configyaml_class(config_text=config_text, context=self.context)
        self.errors = self.loader.errors

    def is_valid(self):
        return not self.errors

    def as_dict(self):
        return self.loader.as_dict()



from __future__ import unicode_literals
import sys

from configyaml.config import AbstractNode
from configyaml.config import DictNode
from configyaml.config import ListNode
from configyaml.loader import ConfigLoader
from configyaml.config import WildcardDictNode
from configyaml.config import TypelessNode


class DummyConfig(WildcardDictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            '*': {
                'class': TypelessNode,
            },
        }
        super(DummyConfig, self).__init__(*args, **kwargs)


class DummyLoader(ConfigLoader):
    config_root_class = DummyConfig


def test_valid_yaml():
    text = """
string_var: stringy
int_var: 8
float_var: 4.5
bool_var: yes
bool_var2: True
dict_var:
  subdict_var: true
list_var:
- a
- false
    """

    # other things you can do in yaml? http://pyyaml.org/wiki/PyYAMLDocumentation#Tokens

    config = DummyLoader(text)

    assert config.is_valid()

    assert config['string_var'] == 'stringy'
    assert config['int_var'] == 8
    assert config['float_var'] == 4.5
    assert config['bool_var'] == True
    assert config['bool_var2'] == True
    assert config['dict_var'] == {'subdict_var': True}
    assert config['list_var'] == ['a', False]


def test_invalid_python_type():
    text = """
  string_var: stringy
  int_var: 8
  float_var: 4.5
  python_str_var: !!python/str "test"
    """

    config = DummyLoader(text)

    assert not config.is_valid()
    assert config.errors[0].description == "could not determine a constructor for the tag 'tag:yaml.org,2002:python/str'"


def test_invalid_line():
    text = """line_a: True
line_b: !2
line_c: 8"""

    config = DummyLoader(text)

    assert not config.is_valid()
    assert len(config.errors) == 1

    error = config.errors[0]
    assert error.title == 'Basic YAML parsing error'
    assert error.description == "could not determine a constructor for the tag '!2'"
    assert error.line == 1  # 2
    assert error.column == 8  # 9

    assert config.as_text() == """line_a: True
line_b: !2
# line_b: !2
#         ^
# --------
# Basic YAML parsing error
# - could not determine a constructor for the tag '!2'
# --------
line_c: 8"""

    # We want to make sure we're getting unicode back
    if sys.version_info.major < 3:
        assert isinstance(config.as_text(), unicode)
    else:
        assert isinstance(config.as_text(), str)

    assert config.as_dict() == {'config': None,
                                'config_text': 'line_a: True\nline_b: !2\nline_c: 8',
                                'errors': [{'description': "could not determine a constructor for the tag '!2'",
                                            'end_column': None,
                                            'end_line': None,
                                            'start_column': 8,
                                            'start_line': 1,
                                            'title': 'Basic YAML parsing error'}],
                                }

def test_bytes_in_unicode_out():
    """Test to make sure that even if we put in bytes that we get out uniform utf8 text, original lines as
       well as error lines.  We've had errors where the output in Python3 has the original lines prefixed
       with 'b' and we want to make sure that doesn't happen again."""
    text = """line_a: True
line_b: !2
line_c: 8"""

    config = DummyLoader(text.encode('utf-8'))

    assert not config.is_valid()

    # We want to make sure we're getting unicode back
    if sys.version_info.major < 3:
        assert isinstance(config.as_text(), unicode)
    else:
        assert isinstance(config.as_text(), str)

    assert config.as_text() == """line_a: True
line_b: !2
# line_b: !2
#         ^
# --------
# Basic YAML parsing error
# - could not determine a constructor for the tag '!2'
# --------
line_c: 8"""




def test_empty_yaml():
    text = """

    """
    config = DummyLoader(text)

    assert not config.is_valid()
    assert len(config.errors) == 1

    error = config.errors[0]
    assert error.title == 'YAML is empty'
    assert error.description == "Your configuration file appears to be empty."
    assert not error.line
    assert not error.column


# can you use mutiple docs? http://pyyaml.org/wiki/PyYAMLDocumentation#LoadingYAML


class StringType(AbstractNode):
    def __init__(self, *args, **kwargs):
        self._type = str
        super(StringType, self).__init__(*args, **kwargs)


class ProjectDictNode(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'github': {
                'class': StringType,
                'required': True,
            },
        }
        super(ProjectDictNode, self).__init__(*args, **kwargs)


class ProjectListNode(ListNode):
    def __init__(self, *args, **kwargs):
        self._list_items_class = ProjectDictNode
        self._min_items_required = 1
        super(ProjectListNode, self).__init__(*args, **kwargs)


class BaseConfig(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'projects': {
                'class': ProjectListNode,
                'required': True,
            },
            'notifications': {
                'class': ProjectListNode,
                'required': True,
            },
        }
        super(BaseConfig, self).__init__(*args, **kwargs)


class ProjectLoader(ConfigLoader):
    config_root_class = BaseConfig


def test_multiple_invalid_lines():
    text = """\
deadbeef: deadbeef
alist:
    - one
    - two
    - three"""

    config = ProjectLoader(text)
    assert not config.is_valid()

    expected_error = """\
deadbeef: deadbeef
# deadbeef: deadbeef
# ^
# --------
# Invalid key
# - Available fields are: notifications, projects
# Required field is missing
# - notifications is a required field
# - projects is a required field
# --------
alist:
# alist:
# ^^^^^
# --------
# Invalid key
# - Available fields are: notifications, projects
# --------
    - one
    - two
    - three"""
    assert expected_error == config.as_text()



from __future__ import unicode_literals

from configyaml.loader import ConfigLoader
from configyaml.config import ChoiceNode


class DummyConfig(ChoiceNode):
    def __init__(self, *args, **kwargs):
        self._choices = ['foo', 'bar']
        super(DummyConfig, self).__init__(*args, **kwargs)


class DummyLoader(ConfigLoader):
    config_root_class = DummyConfig


def test_valid_choice():
    value = 'foo'
    loader = DummyLoader(value)
    assert loader.is_valid()


def test_valid_choice_whitespace():
    value = ' foo '
    loader = DummyLoader(value)
    assert loader.is_valid()


def test_valid_choice_2():
    value = 'bar'
    loader = DummyLoader(value)
    assert loader.is_valid()


def test_invalid_choice():
    value = 'deadbeef'
    loader = DummyLoader(value)
    assert not loader.is_valid()



from ..errors import ConfigError


class AbstractNode(object):
    """Configuration Base Class

    All nodes in the configuration inherit from this base class
    to define types of nodes and groups of nodes; it is not intended
    to be used directly, but rather to be subclassed.

    """
    def __init__(self, value, value_node=None, context={}, variables={}, key=None, parent=None):
        """Initialize a node

        :param value:
        :param value_node:
        :param context:
        :param variables:
        :param key:
        :param parent:
        """
        self._original_value = value
        self._value_node = value_node  # yaml node obj
        self._context = context
        self._variables = variables
        self._key = key
        self._parent = parent
        self._errors = []

        self._is_variable = isinstance(self._original_value, str) and self._original_value.startswith('$')

        self._value = self._render_value(self._original_value)

        if self._errors:
            # if we already have errors, then just quit
            return

        self._validate()

        # probably a nicer way to do this by breaking down the steps
        # (creating all objects, setting context, then validating) but turned out to be a mess in 1st try
        if self._context_to_inject():
            self._context.update(self._context_to_inject())
            # then we need to revalidate with that context
            self._validate()

    def is_valid(self):  # type: () -> bool
        """Tests whether this node is valid (but not descendants)

        :rtype: bool
        """
        return not self._errors

    def _render_value(self, value):
        if self._is_variable:
            if not isinstance(self._variables, dict):
                raise TypeError('variables must be a dict')

            variable_name = value[1:]

            if variable_name in self._variables:
                return self._variables[variable_name]
            else:
                self._add_error(
                    title='Variable not found',
                    # it is safe to give the keys of the variables, but not values
                    description='\'{}\' was not found in {}'.format(variable_name, sorted(list(self._variables.keys())))
                )

        elif isinstance(value, str) and value.startswith('\$'):
            # strip the leading \, which told us not to treat it as a variable
            # but presumably they want the string without that character
            return value[1:]

        return value

    def _key_name(self):  # type: () -> str
        """Return the key referring to this object

        The default value is the lower case version of the class name

        :rtype: str
        """
        if self._key is not None:
            return self._key

        return self.__class__.__name__.lower()

    def __nonzero__(self):  # type: () -> bool
        """Python 2 support for object truth

        :rtype: bool
        """
        return True

    def __bool__(self):  # type: () -> bool
        """Python 3 support for object truth

        :rtype: bool
        """
        return True

    def _path(self):  # type: () -> str
        """Return the dotted path representation of this object

        :rtype: str
        """
        if self._parent:
            return '{}.{}'.format(self._parent._path(), self._key_name())

        return self._key_name()

    def _add_error(self, *args, **kwargs):  # type: () -> None
        """Convenience function to add an error to this object, with line numbers

        An error title or description should not accidentally leak self._value, for privacy/redaction purposes.

        :rtype: None
        """
        if kwargs.get('node', None):
            # if node specified and not none
            error = ConfigError.create_from_yaml_node(
                *args,
                **kwargs
            )
        elif self._value_node:
            # default to using the node if we have one
            error = ConfigError.create_from_yaml_node(
                node=self._value_node,
                *args,
                **kwargs
            )
        else:
            # no nodes or error_obj to attach
            error = ConfigError(*args, **kwargs)

        self._errors.append(error)

    def _get_descendants_errors(self):  # type: () -> List(ConfigError)
        """Recursively get errors from descendants

        :rtype: list of :ConfigErrors:"""
        descendants_errors = []
        if hasattr(self, '_children'):
            if isinstance(self._children, (list, tuple)):
                for c in self._children:
                    descendants_errors += c._get_all_errors()
            elif isinstance(self._children, dict):
                for c in self._children.values():
                    descendants_errors += c._get_all_errors()

        return descendants_errors

    def _get_all_errors(self):  # type: () -> List(ConfigError)
        """Collect and return a list of all errors

        Collects all errors from this node and all of its descendants.

        :rtype: list of :ConfigErrors:
        """
        return self._errors + self._get_descendants_errors()

    def _validate(self):  # type: () -> None
        """Run validation, save errors to object in self._errors"""
        # class can specify it's empty obj -- list would have empty of []
        self._errors = []

        self._validate_type()

        if self.is_valid():
            self._validate_value()

    def _validate_type(self):  # type: () -> None
        """Validation to ensure value is the correct type"""
        if not isinstance(self._value, self._type):
            title = '{} has an invalid type'.format(self._key_name())
            description = '{} must be a {}'.format(self._key_name(), self._type.__name__)

            self._add_error(title=title, description=description)

    def _validate_value(self):  # type: () -> None
        """Do custom validation here, only runs if valid up to this point"""
        pass

    def _context_to_inject(self):  # type: () -> dict
        """An optional dictionary of context to be injected into children"""
        return {}

    def _as_dict_to_inject(self, redact=False):  # type: () -> dict
        """Additional fields to inject into as_dict"""
        return {}

    def _as_dict(self, redact=False):  # type: () -> dict
        if redact and self._should_redact():
            return self._as_redacted_dict()

        d = {}
        d['value'] = self._value
        if self._errors:
            d['errors'] = [x.as_dict() for x in self._errors]
        d.update(self._as_dict_to_inject(redact=redact))
        return d

    def _should_redact(self):
        return self._is_variable and self._value != self._original_value

    def _as_redacted_dict(self):
        d = {}

        d['value'] = '[REDACTED]'
        d['redacted'] = True

        if self._errors:
            d['errors'] = [x.as_dict() for x in self._errors]

        d.update(self._as_dict_to_inject(redact=True))

        return d



from .base import AbstractNode


class DictNode(AbstractNode):
    def __init__(self, *args, **kwargs):
        self._type = dict
        self._children = {}
        super(DictNode, self).__init__(*args, **kwargs)

    def _validate_value(self):
        self._validate_required_keys()
        self._set_default_fields()

        for key, value in self._value.items():

            if key in self._dict_fields:
                # get class for key
                field_class = self._dict_fields[key]['class']
                field = field_class(
                    value=self._value[key],
                    value_node=self._find_node_for_key_value(key),
                    context=self._context,
                    variables=self._variables,
                    key=key,
                    parent=self
                )

                # set self.FIELD_NAME so we can get children directly
                # these will only be keys we specify, so should be safe names
                self.__dict__[key] = field
                self._children[key] = field
            else:
                # cannot use key
                self._add_error(
                    node=self._find_node_for_key(key),
                    title='Invalid key',
                    description='Available fields are: {}'.format(', '.join(sorted(self._dict_fields.keys()))),
                )

    def _validate_required_keys(self):
        required_keys = [k for k, v in self._dict_fields.items() if v.get('required', False)]
        for k in required_keys:
            if k not in self._value:
                self._add_error(
                    title='Required field is missing',
                    description='{} is a required field'.format(k)
                )

    def _set_default_fields(self):
        for k, v in self._dict_fields.items():
            if 'default' in v:
                default = v['default']
                instance = v['class'](
                    value=default,
                    context=self._context,
                    variables=self._variables,
                    key=k,
                    parent=self
                )
                self.__dict__[k] = instance
                self._children[k] = instance

    def _find_node_for_key_value(self, key):
        if not self._value_node:
            return None

        for children in self._value_node.value:
            # value would be the second ScalarNode
            if len(children) > 1:
                key_node = children[0]
                if key_node.value == key:
                    return children[1]  # return the following value node

        return None

    def _find_node_for_key(self, key):
        if not self._value_node:
            return None

        for children in self._value_node.value:
            # key would be the first ScalarNode
            key_node = children[0]
            if key_node.value == key:
                return key_node

        return None

    def __getitem__(self, key):
        return self._children[key]

    def __len__(self):
        return len(self._children)

    def _as_dict(self, redact=False):
        if redact and self._should_redact():
            return self._as_redacted_dict()

        d = {}
        for k in self._dict_fields.keys():
            d[k] = self[k]._as_dict(redact=redact)

        if self._errors:
            d['errors'] = [x.as_dict() for x in self._errors]

        d.update(self._as_dict_to_inject(redact=redact))

        return d



from .base import AbstractNode
from .list import ListNode
from .dict import DictNode
from .choice import ChoiceNode
from .nodes import (
    IntegerNode,
    PositiveIntegerNode,
    BoolNode,
    StringNode,
    WildcardDictNode,
    TypelessNode,
    RegexNode,
)



# everything here should import correctly

from configyaml.loader import ConfigLoader
from configyaml.validator import ConfigValidator

from configyaml.config import AbstractNode
from configyaml.config import ListNode
from configyaml.config import DictNode
from configyaml.config import ChoiceNode
from configyaml.config import IntegerNode
from configyaml.config import PositiveIntegerNode
from configyaml.config import BoolNode
from configyaml.config import StringNode
from configyaml.config import WildcardDictNode
from configyaml.config import RegexNode



import yaml

from .base import AbstractNode


class ListNode(AbstractNode):
    _list_item_class = None

    def __init__(self, *args, **kwargs):
        self._type = list
        self._children = []  # so we can append, etc

        if not hasattr(self, '_min_items_required'):
            # by default, don't have to have anything
            # set to 1 in custom __init__ to require at least 1 item
            self._min_items_required = 0

        if not self._list_item_class:
            raise AttributeError('_list_item_class must be defined in subclasses of ListNode')

        super(ListNode, self).__init__(*args, **kwargs)

    def _validate_value(self):
        if len(self._value) < self._min_items_required:
            self._add_error(
                title='Minimum items requirement not met',
                description='Must have at least {} item(s)'.format(self._min_items_required)
            )

        for index, item in enumerate(self._value):
            field_class = self._list_item_class
            field = field_class(
                value=item,
                value_node=self._find_node_for_list_index(index),
                context=self._context,
                variables=self._variables,
                parent=self,
                key=index
            )

            self._children.append(field)

    def _find_node_for_list_index(self, index):
        if not self._value_node:
            return None

        if not isinstance(self._value_node, yaml.nodes.SequenceNode):
            # the original node was not a list
            # - could have been a variable string
            #
            # return the original node so that the yaml text can place the error there
            return self._value_node

        return self._value_node.value[index]

    def __getitem__(self, key):
        return self._children[key]

    def __len__(self):
        return len(self._children)

    def _as_dict(self, redact=False):
        if redact and self._should_redact():
            return self._as_redacted_dict()

        d = {
            'items': [x._as_dict(redact=redact) for x in self._children],
        }

        if self._errors:
            d['errors'] = [x.as_dict() for x in self._errors]

        d.update(self._as_dict_to_inject(redact=redact))
        return d



from __future__ import unicode_literals

from configyaml.config import DictNode
from configyaml.config import AbstractNode


class DummyFoo(AbstractNode):
    def __init__(self, *args, **kwargs):
        self._type = str
        super(DummyFoo, self).__init__(*args, **kwargs)


class DummyConfig(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyFoo,
            }
        }
        super(DummyConfig, self).__init__(*args, **kwargs)


class DummyConfigRequired(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyFoo,
                'required': True,
            }
        }
        super(DummyConfigRequired, self).__init__(*args, **kwargs)


class DummyConfigDefault(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyFoo,
                'default': 'test',
            }
        }
        super(DummyConfigDefault, self).__init__(*args, **kwargs)


class DummyConfigDefaultNone(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyFoo,
                'required': False,
                'default': None,
            }
        }
        super(DummyConfigDefaultNone, self).__init__(*args, **kwargs)


def test_normal():
    value = {}
    config = DummyConfig(value=value)
    assert config.is_valid()
    assert 'foo' not in config._children
    assert not hasattr(config, 'foo')

    value = {'foo': 'bar'}
    config = DummyConfig(value=value)
    assert config.is_valid()
    assert 'foo' in config._children
    assert hasattr(config, 'foo')


def test_invalid_key():
    value = {'bar': 'baz'}
    config = DummyConfig(value=value)
    assert not config.is_valid()
    assert config._errors[0].title == 'Invalid key'
    assert not hasattr(config, 'bar')


def test_invalid_value_key_name():
    """
    An error in a dict field should know the name it was used as,
    and not use the name.lower() of its class (dummyfoo)
    """
    value = {'foo': 2}
    config = DummyConfig(value=value)
    assert config.is_valid()  # the dict itself is valid
    assert not config.foo.is_valid()
    assert config.foo._errors[0].description == 'foo must be a str'


def test_required():
    value = {}
    config = DummyConfigRequired(value=value)
    assert not config.is_valid()
    assert config._errors[0].title == 'Required field is missing'

    value = {'foo': 'bar'}
    config = DummyConfigRequired(value=value)
    assert config.is_valid()


def test_default():
    value = {}
    config = DummyConfigDefault(value=value)
    assert config.is_valid()
    assert config.foo._value == 'test'
    assert config.foo._parent == config
    assert config.foo._key == 'foo'

    value = {'foo': 'bar'}
    config = DummyConfigDefault(value=value)
    assert config.is_valid()
    assert config.foo._value == 'bar'
    assert config.foo._parent == config
    assert config.foo._key == 'foo'


def test_get_item():
    value = {'foo': 'bar'}
    config = DummyConfig(value=value)
    assert config.is_valid()
    assert config['foo']._value == 'bar'


def test_default_none():
    value = {}
    config = DummyConfigDefaultNone(value=value)
    assert config.is_valid()
    assert config.foo._value == None

    value = {'foo': 'bar'}
    config = DummyConfigDefaultNone(value=value)
    assert config.is_valid()
    assert config.foo._value == 'bar'



from __future__ import unicode_literals

import pytest

from configyaml.config import ListNode
from configyaml.loader import ConfigLoader
from configyaml.config import DictNode
from configyaml.config import AbstractNode


class DummyFoo(AbstractNode):
    def __init__(self, *args, **kwargs):
        self._type = str
        super(DummyFoo, self).__init__(*args, **kwargs)


class DummyConfig(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyFoo,
            }
        }
        super(DummyConfig, self).__init__(*args, **kwargs)


class DummyLoader(ConfigLoader):
    config_root_class = DummyConfig


class DummyComplexConfig(DictNode):
    def __init__(self, *args, **kwargs):
        self._dict_fields = {
            'foo': {
                'class': DummyMinConfigListNode,
            },
            'bar': {
                'class': DummyFoo,
            }
        }
        super(DummyComplexConfig, self).__init__(*args, **kwargs)


class DummyMinConfigListNode(ListNode):
    def __init__(self, *args, **kwargs):
        self._min_items_required = 1
        self._list_item_class = DummyFoo
        super(DummyMinConfigListNode, self).__init__(*args, **kwargs)


class DummyComplexLoader(ConfigLoader):
    config_root_class = DummyComplexConfig


def test_dict_error_propogation():
    value = """foo: ['list']"""
    loader = DummyLoader(value)
    assert not loader.is_valid()
    assert len(loader.errors) == 1

    assert len(loader.config_root._errors) == 0
    assert len(loader.config_root._get_descendants_errors()) == 1
    assert len(loader.config_root._get_all_errors()) == 1

    assert len(loader.config_root.foo._errors) == 1
    assert len(loader.config_root.foo._get_descendants_errors()) == 0
    assert len(loader.config_root.foo._get_all_errors()) == 1


def test_valid_as_dict():
    value = "{ 'foo': 'bar'}"
    loader = DummyLoader(value)
    assert loader.is_valid()
    assert loader.as_dict()['config'] == {'foo': {'value': 'bar'}}


def test_valid_as_text():
    value = "{ 'foo': 'bar'}"
    loader = DummyLoader(value)
    assert loader.is_valid()
    assert loader.as_text() == value


def test_as_dict_with_lists():
    value = "{ 'foo': ['x', 'y'], 'bar': 'deadbeef'}"
    loader = DummyComplexLoader(value)
    assert loader.is_valid()
    assert loader.as_dict()['config'] == {'foo': {'items': [{'value': 'x'}, {'value': 'y'}]}, 'bar': {'value': 'deadbeef'}}
# def test_list_error_propogation():


def test_asserts_on_invalid_subclass():
    class InvalidLoader(ConfigLoader):
        pass

    value = ['a', 'b']
    with pytest.raises(AttributeError):
        InvalidLoader(value)



from __future__ import unicode_literals

import pytest

from configyaml.config import ListNode
from configyaml.config import AbstractNode


class DummyFoo(AbstractNode):
    def __init__(self, *args, **kwargs):
        self._type = str
        super(DummyFoo, self).__init__(*args, **kwargs)


class DummyConfigListNode(ListNode):
    def __init__(self, *args, **kwargs):
        self._list_item_class = DummyFoo
        super(DummyConfigListNode, self).__init__(*args, **kwargs)


class DummyMinConfigListNode(ListNode):
    def __init__(self, *args, **kwargs):
        self._min_items_required = 1
        self._list_item_class = DummyFoo
        super(DummyMinConfigListNode, self).__init__(*args, **kwargs)


def test_list_empty():
    value = []
    config = DummyConfigListNode(value=value)
    assert config.is_valid()
    assert len(config) == 0

def test_list_none():
    value = None
    config = DummyConfigListNode(value=value)
    assert len(config) == 0

def test_list_empty_with_minimum():
    value = []
    config = DummyMinConfigListNode(value=value)
    assert not config.is_valid()
    assert config._errors[0].title == 'Minimum items requirement not met'

def test_list_with_two():
    value = ['a', 'b']
    config = DummyConfigListNode(value=value)
    assert config.is_valid()
    assert len(config) == 2

def test_list_attr():
    value = ['a', 'b']
    config = DummyConfigListNode(value=value)
    assert config[0]

def test_assert_on_missing_list_item_class():
    class InvalidConfigListNode(ListNode):
        pass

    value = ['a', 'b']
    with pytest.raises(AttributeError):
        config = InvalidConfigListNode(value=value)



import re

from .base import AbstractNode
from .dict import DictNode


class WildcardDictNode(DictNode):
    """A dictionary node where nearly any key is valid

       The only invalid keys are:
       - an actual asterisk '*'
       - a key that begins with '_'
       """
    def _validate_value(self):
        for key, value in self._value.items():
            # where key name doesn't matter (ex. groups)
            key_valid, explanation = self._key_name_is_valid(key)
            if key_valid:
                field_class = self._dict_fields['*']['class']
                field = field_class(
                    value=value,
                    value_node=self._find_node_for_key_value(key),
                    context=self._context,
                    variables=self._variables,
                    key=key,
                    parent=self
                )

                # don't set __dict__ if they can use any key
                self._children[key] = field
            else:
                self._add_error(
                    node=self._find_node_for_key(key),
                    title='Invalid field name',
                    description=explanation
                )

    def _as_dict(self, redact=False):
        if redact and self._should_redact():
            return self._as_redacted_dict()

        d = {}
        for group_name in self._children.keys():
            d[group_name] = self[group_name]._as_dict(redact=redact)

        if self._errors:
            d['errors'] = [x.as_dict() for x in self._errors]

        d.update(self._as_dict_to_inject(redact=redact))

        return d

    def _key_name_is_valid(self, key):
        if key == '*':
            return False, 'Field name cannot be "*"'

        if key.startswith('_'):
            return False, 'Cannot start field name with a "_"'

        return True, 'Valid'


class BoolNode(AbstractNode):
    """A node that must validate as a bool"""
    def __init__(self, *args, **kwargs):
        self._type = bool
        super(BoolNode, self).__init__(*args, **kwargs)


class StringNode(AbstractNode):
    """A node that must validate as a string"""
    def __init__(self, *args, **kwargs):
        self._type = str
        super(StringNode, self).__init__(*args, **kwargs)


class RegexNode(StringNode):
    """A node that must validate as a regular expression"""
    def __init__(self, *args, **kwargs):
        self.regex = None
        super(RegexNode, self).__init__(*args, **kwargs)

    def _validate_value(self):
        try:
            self.regex = re.compile(self._value)
        except re.error as e:
            self._add_error(title='Invalid regex', description=str(e))


class IntegerNode(AbstractNode):
    """A node that must validate as an integer"""
    def __init__(self, *args, **kwargs):
        self._type = int
        super(IntegerNode, self).__init__(*args, **kwargs)


class PositiveIntegerNode(IntegerNode):
    """A node that must validate as a positive integer"""
    def _validate_value(self):
        if self._value < 0:
            self._add_error(title="Invalid Value", description="Must be a positive integer")


class TypelessNode(AbstractNode):
    """A node that does not have to validate as any specific type"""
    def _validate_type(self):
        pass



from __future__ import unicode_literals
import yaml

from .errors import ConfigError


class ConfigLoader(object):
    config_root_class = None

    def __init__(self, config_text, context={}, variables={}):
        if not self.config_root_class:
            raise AttributeError('config_root_class must defined in subclasses of ConfigLoader')

        self.config_text = config_text
        self.config_dict = None
        self.config_root = None
        self.context = context
        self.variables = variables

        self._errors = []
        self.load()

    def load(self):
        """
        - should not be empty
        - yaml itself should be valid (should we grab more than 1 yaml error at a time?)
        - we should have whitelist of fields the can set at each level, and start
          getting objects out of those, processing grammar if necessary, validating other settings
          errors coming all the way back up
        """
        if not isinstance(self.config_text, str):
            self.config_text = self.config_text.decode('utf-8')

        if not self.config_text.strip():
            self._errors.append(ConfigError(title='YAML is empty', description='Your configuration file appears to be empty.'))
            return

        # simple way to check that yaml itself is valid
        try:
            self.config_dict = yaml.safe_load(self.config_text)
        except yaml.YAMLError as e:
            error = ConfigError.create_from_yaml_error(e)
            self._errors.append(error)
            # could have more than 1 line, keep giong

        if self.config_dict:
            # we have valid yaml with data, so start checking the components
            node_tree = yaml.compose(self.config_text)
            # give it the parsed settings, and the node info
            self.config_root = self.config_root_class(
                value=self.config_dict,
                value_node=node_tree,
                context=self.context,
                variables=self.variables,
            )

    @property
    def errors(self):
        if self.config_root:
            return self._errors + self.config_root._get_all_errors()

        return self._errors

    def is_valid(self):
        return not self.errors

    def __getitem__(self, key):
        # just pass off as dict right now
        return self.config_dict[key]

    def as_dict(self, redact=False):
        d = {
            'config_text': self.config_text,
            'config': self.config_root._as_dict(redact=redact) if self.config_root else None,
        }
        if self.errors:
            d['errors'] = [x.as_dict() for x in self.errors]
        return d

    def as_text(self):
        if self.is_valid():
            return self.config_text

        else:
            output = []
            errored_lines = set([x.line for x in self.errors])

            for index, line in enumerate(self.config_text.splitlines()):
                output.append(line)
                if index in errored_lines:
                    errors_on_line = [x for x in self.errors if x.line == index]
                    # If there is more than one error on a line, try to group them by title and place in a
                    # single comment block
                    if len(errors_on_line) > 1:
                        error_str = """# {line}\n# ^\n# --------\n""".format(line=line)
                        unique_titles = set([x.title for x in errors_on_line])

                        for t in sorted(unique_titles):
                            error_str += """# {title}\n""".format(title=t)
                            for d in sorted([x.description for x in errors_on_line if x.title == t]):
                                error_str += """# - {description}\n""".format(description=d)

                        error_str += """# --------"""
                    else:
                        e = errors_on_line[0]
                        num_markers = 1 if not e.end_column else e.end_column - e.start_column
                        markers = '^' * num_markers
                        error_str = """\
# {line}
# {markers}
# --------
# {title}
# - {description}
# --------""".format(           line=line,
                            markers=markers.rjust(e.start_column + 1),
                            title=e.title,
                            description=e.description,
                        )
                    output.append(error_str)

            text = '\n'.join([str(x) for x in output])
            return text
